import logging
import time
from queue import Queue

from astrometry_net_client.session import Session
from astrometry_net_client.settings import Settings
from astrometry_net_client.uploads import FileUpload

log = logging.getLogger(__name__)

# TODO: document this value.
MAX_WORKERS = 10


class Client:
    """
    Higher level class which makes the interaction with the API easy.

    Parameters
    ----------
    session: :py:class:`astrometry_net_client.session.Session`
        Optional argument to provide the session for the API. If not provided,
        you must either specify the keyword ``api_key`` or ``location``, or
        have the key present in your environment. See
        :py:class:`astrometry_net_client.session.Session`
    settings: :py:class:`astrometry_net_client.settings.Settings`, dict
        An optional settings object, which will be applied to all uploads.
        If not specified, the settings object will be created from given
        keyword arguments which correspond to valid settings. For this see
        :py:class:`astrometry_net_client.settings.Settings`.
    kwargs: arguments
        Used to create a session or settings object, if either is not
        specified. Will extract the relevant arguments relevant to the object
        which is created before passing them to the object constructor.
        See :py:class:`astrometry_net_client.settings.Settings` and
        :py:class:`astrometry_net_client.session.Session`
    """

    def __init__(self, session=None, settings=None, **kwargs):
        # TODO, make session optional?
        if session is None:
            args = {k: v for k, v in kwargs.items() if k in {"api_key", "key_location"}}
            self.session = Session(**args)
        else:
            self.session = session

        if settings is None:
            setting_args = {k: v for k, v in kwargs.items() if k in Settings._settings}
            self.settings = Settings(**setting_args)
        else:
            self.settings = Settings(settings)

        log.info("Logging in")
        self.session.login()

    def upload_files_gen(
        self,
        files_iter,
        queue_size=MAX_WORKERS,
    ):
        """
        Generator which uploads a number of files concurrently, yielding the
        :py:class:`astrometry_net_client.statusables.Job` & filename when done.

        Note that the ``files_iter`` argument is fully consumed when submitting
        to the executor is done. This means that the results will only be
        yielded once the iterator is consumed.

        Parameters
        ----------
        files_iter: iterable
            Some iterable containing paths to the files which will be uploaded.
            Is fully consumed before any result is yielded.
        queue_size: int, optional
            A positive integer, controlling the size of the queue. This will
            determine the maximum number of simultaneous submissions. Must be
            greater than 0 and lower than :py:const:`MAX_WORKERS`. Default is
            :py:const:`MAX_WORKERS`.

        Yields
        ------
        (:py:class:`astrometry_net_client.statusables.Job`, ``str``):
            A tuple containing the finished
            :py:class:`astrometry_net_client.statusables.Job` and the
            corresponding filename. Yields when the Job is finished.
            NOTE: Order of yielded filenames can (and likely will) be different
            from the given ``files_iter``

        Raises
        ------
        ValueError
            When the queue_size is invalid.
        """
        SLEEP_TIME = 0.3  # seconds

        files_iter = iter(files_iter)
        if queue_size < 1 or queue_size > MAX_WORKERS:
            raise ValueError(
                "queue_size must be greater than 0 and less or equal to ",
                f"{MAX_WORKERS}, was: {queue_size}",
            )
        processing_queue = Queue(maxsize=queue_size)

        # Populate queue initially
        for _, filename in zip(range(queue_size), files_iter):
            self._insert_submission(filename, processing_queue)

        while not processing_queue.empty():
            filename, submission, job = processing_queue.get()
            log_msg = "Checking file {}, job exists: {}"
            log.debug(log_msg.format(filename, job is not None))
            # The item in the queue has 2 states; if it is still only a
            # submission job will be None and we have to create a job out of
            # it. When the job is made, we can check if the job is done. When
            # the job is finished return (yield) the value, otherwise put it
            # back in the queue.

            if job is None:
                submission.status()
                if submission.done():
                    job = submission.jobs[0]
                else:
                    processing_queue.put((filename, submission, job))
                    continue

            job.status()
            if job.done():
                try:
                    filename = next(files_iter)
                except StopIteration:
                    pass
                else:
                    self._insert_submission(filename, processing_queue)
                log_msg = "FINISHED submission {}, yielding..."
                log.info(log_msg.format(filename))
                yield (job, filename)
            else:
                processing_queue.put((filename, submission, job))

            time.sleep(SLEEP_TIME)

    def _insert_submission(self, filename, queue):
        """
        Helper function which creates an upload for the given filename, and
        inserts the submission in a queue. This is not intended to be used
        by a user.

        Parameters
        ----------
        filename: str
            The filename of the file to be submitted into the queue.
        queue: Queue
            The queue in which to insert the submission.
        """
        log_msg = "Submitting file {}"
        log.info(log_msg.format(filename))
        upl = FileUpload(filename, session=self.session, settings=self.settings)
        submission = upl.submit()
        queue.put((filename, submission, None))

    def upload_file(self, filename, settings=None):
        """
        Uploads file and returns completed job when finished solving.

        Parameters
        ----------
        filename: str
            Location + name of the file to upload.
        settings: :py:class:`astrometry_net_client.settings.Settings`
            An optional settings dict which only applies to this specific
            upload. Will override the default settings with which the
            :py:class:`Client` was constructed.

        Returns
        -------
        Job: :py:class:`astrometry_net_client.statusables.Job`
            The job of the resulting upload. NOTE: It is possible that the job
            did not succeed, therefore check with
            :py:meth:`astrometry_net_client.statusables.Statusable.success` if
            it did.
        """
        upl_settings = self.settings
        if settings is not None:
            upl_settings.update(settings)

        upl = FileUpload(filename, session=self.session, settings=upl_settings)
        start = time.time()
        submission = upl.submit()

        msg = "File {} submitted, waiting for it to finish"
        log.info(msg.format(filename))

        submission.until_done()  # blocks here

        # pretty much guarenteed to have exactly one job
        job = submission.jobs[0]

        job.until_done()  # blocks here
        end = time.time()

        msg = "Processing of file {} finished in {}s"
        log.info(msg.format(filename, end - start))

        return job

    def calibrate_file_wcs(self, filename, settings=None):
        """
        Uploads a file, returning the wcs header if it succeeds.

        Parameters
        ----------
        filename: str
            Location + name of the file to upload.
        settings: :py:class:`astrometry_net_client.settings.Settings`
            An optional settings dict which only applies to this specific
            upload. Will override the default settings with which the
            :py:class:`Client` was constructed.

        Returns
        -------
        :py:class:`astropy.io.fits.Header` or None
            Returns the Header if successful, None otherwise.
        """
        job = self.upload_file(filename, settings=settings)

        if job.success():
            return job.wcs_file()
        return None



import logging
import os
from typing import Optional, cast

from astrometry_net_client.config import login_url, read_api_key
from astrometry_net_client.exceptions import (
    APIKeyError,
    InvalidRequest,
    InvalidSessionError,
    LoginFailedException,
)
from astrometry_net_client.request import PostRequest, Request

log = logging.getLogger(__name__)


class Session(object):
    """
    Class to hold information on the Astrometry.net API session. Is able to
    read an API key from multiple possible locations, and login to retrieve
    a session key (stored in the id attribute).

    Attributes
    ----------
    logged_in : bool
        Boolean indicating if a :py:func:`login` has happened successfully
        since creation of the class. Does not guarentee that the session
        :py:attr:`key` is (still) valid.
    api_key : str
        The API key used to log in.
    key : str
        The session key (not the :py:attr:`api_key`) which is given by the API
        after a valid :py:func:`login`.

    Examples
    --------
    Normal usage of a session is as follows; you create the Session class, and
    then :py:func:`login`:

    >>> session = Session('the api key')
    >>> session.login()

    There are multiple ways of giving your api key. You can just give it as
    a plain string (not very safe):

    >>> session_1 = Session(api_key='the api key')

    You can enter it in a separate private file and give the path to it. This
    file should only contain the key, nothing else.

    >>> session_2 = Session(key_location='path/to/file/with/key')

    You can specify the key as an environment variable named
    ``ASTROMETRY_API_KEY``.

    >>> session_3 = Session()

    If you do not specify any of the above, an exception will be thrown:

    >>> Session()
    APIKeyError

    Raises
    ------
    :py:exc:`APIKeyError`
        When no API key is specified (see examples)
    """

    url: str = login_url

    def __init__(
        self, api_key: Optional[str] = None, key_location: Optional[str] = None
    ):
        if api_key is not None:
            self.api_key = api_key.strip()
        elif key_location is not None:
            self.api_key = read_api_key(key_location)
        else:
            env_key = os.environ.get("ASTROMETRY_API_KEY")
            if env_key is None:
                raise APIKeyError(
                    "No api key found or given. "
                    "Specify an API key using one of: "
                    "1. A direct argument, "
                    "2. A location of a file containing the key, "
                    "3. An environment variable named ASTROMETRY_API_KEY."
                )
            self.api_key = cast(str, env_key)

        self.logged_in = False

    def login(self, force: bool = False) -> None:
        """
        Method used to log-in or start a session with the Astrometry.net API.

        Only sends a request if it is absolutely needed (or forced to).  If
        :py:attr:`logged_in` is True, it assumes the session is still valid
        (there is no way to check if the session is still valid).

        After a successful login, the session key is stored in the
        :py:attr:`key` attribute and :py:attr:`logged_in` is set to `True`.

        Parameters
        ----------
        force: bool
            Forces the login by ignoring the :py:attr:`logged_in` boolean.

        Raises
        ------
        LoginFailedException
            If the login response does not have the status 'success'.
        """
        if not force and self.logged_in:
            return

        r = PostRequest(self.url, data={"apikey": self.api_key})
        try:
            response = cast(dict, r.make())
        except InvalidRequest:
            raise LoginFailedException("The api key given is not valid")

        self.key: str = response["session"]
        self.logged_in = True


class SessionRequest(Request):
    """
    Wraps the normal Request around an authentication layout, ensuring the
    user is logged in and the session key is send alongside the request.

    The separate login request (if needed) is only send just before the
    original request is made, (e.g. when calling make / _make_request).

    Attributes
    ----------
    session: :py:class:`Session`
    """

    def __init__(self, session: Session, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = session

    def _make_request(self) -> dict:
        self.session.login()
        self.data["session"] = self.session.key
        try:
            # A login request will always return a dictionary
            return cast(dict, super()._make_request())
        except InvalidSessionError:
            log.info("Session expired, loggin in again")
            self.session.login(force=True)
            # update the session key for the request as well
            self.data["session"] = self.session.key

        # A login request will always return a dictionary
        return cast(dict, super()._make_request())



#!/usr/bin/env python3
import argparse
import logging
import textwrap
from pathlib import Path

from astropy.io import fits

from astrometry_net_client import Client, Settings

# These lines set up logging
FMT = "[%(asctime)s] %(levelname)-8s |" " %(funcName)s - %(message)s"
logging.basicConfig(level=logging.INFO, format=FMT)
log = logging.getLogger(__name__)


def is_fits(filename: Path):
    """
    Boolean function to test if the extension of the filename provided
    is either .fits or .fit (upper- or lowercase).

    Parameters
    ----------
    filename: Path

    Returns
    -------
    bool
    """
    end = filename.suffix.upper()
    return end.endswith(".FITS") or end.endswith(".FIT")


def main():
    # Commandline argument parsing
    args = parse_arguments()

    s = Settings()
    if args.plate_scale:
        s.set_scale_estimate(
            args.plate_scale, args.plate_scale_err, unit="arcsecperpix"
        )
    if args.plate_scale_range:
        s.set_scale_range(
            args.plate_scale_range[0], args.plate_scale_range[1], unit="arcsecperpix"
        )
    if args.fov_width:
        s.set_scale_estimate(args.fov_width, args.fov_width_err, unit="arcminwidth")
    if args.fov_width_range:
        s.set_scale_estimate(args.fov_width, args.fov_width_err, unit="arcminwidth")

    log.info("Initializing client (logging in)")
    c = Client(api_key=args.key, key_location=args.key_location, settings=s)
    log.info("Log in done")

    # Prepare the output directory
    output_dir = Path(args.output).absolute()
    if output_dir.exists():
        log.warning(f"Output directory: '{output_dir}' already exists")
    else:
        log.info(f"Output directory '{output_dir}' does not exist, creating...")
        output_dir.mkdir(parents=True)

    DO_OVERWRITE = args.overwrite_solve
    if DO_OVERWRITE:
        log.warning(
            "The overwrite-solve is ON, files that exists in the output directory may be overwritten."
        )

    files = [Path(f) for f in args.fits_file]
    # iterate over all the fits files in the specified diretory, ignore all files
    # which are not fits files.
    fits_files = filter(is_fits, files)

    # give the iterable of filenames to the function, which returns a
    # generator, generating pairs containing the finished job and filename.
    result_iter = c.upload_files_gen(fits_files)

    # Iterate over the jobs when they are finished
    for job, filename in result_iter:

        if not job.success():
            log.info("File {} Failed".format(filename))
            continue

        # If the job was successful, we want to get the new wcs file from astrometry.net
        wcs = job.wcs_file()
        # Then we want to add the resulting WCS to the existing file, and write in a
        # new location
        with fits.open(filename) as hdul:
            hdul[0].header.extend(wcs)

            write_filename = output_dir / filename.name

            log.info("Writing to {}...".format(write_filename))
            try:
                hdul.writeto(write_filename, overwrite=DO_OVERWRITE)
            except OSError:
                log.error("File {} already exists.".format(write_filename))


def parse_arguments():
    parser = argparse.ArgumentParser(
        prog="anc_upload.py",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(
            """
                Description
                -----------
                Upload a list of FITS files to find a WCS using nova.astrometry.net, and put the solved files (original file + extended with WCS) in the 'output' directory.
                REQUIRES an account on https://nova.astrometry.net to get an api key. See: https://nova.astrometry.net/api_help

                Setting some basic settings is supported via command line arguments, but for more specific Settings edit the script!

                Examples
                --------
                A very basic example is:
                    $ anc_upload.py --key-location ./key file1.fits file2.fits ./dir_with_files/*

                This assumes you have the api key in the current directory in a file named 'key'.
                It will upload 'file1.fits', 'file2.fits' and all fits files in the 'dir_with_files' directory.

                Another example, setting the plate scale to be between 0.56 and 1.75 arcsec/pixel:
                    $ anc_upload.py --key-location ./key --plate-scale-range 0.56 1.75 file1.fits file2.fits file3.fits
                """
        ),
        epilog="Script and library made by Sten Sipma. For comments and suggestions, please make an issue on Github at https://github.com/StenSipma/astrometry_net_client",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="DIRECTORY",
        default="./anc_output",
        help="Directory in which to put the solved files. Default='./anc_output'",
    )
    parser.add_argument(
        "fits_file", type=str, nargs="+", help="One or more FITS files to upload"
    )
    parser.add_argument(
        "--overwrite-solve",
        action="store_true",
        help="If a file with a similar name already exists in the 'output' directory, overwrite it. Default: False",
    )

    # One of the methods to specify the key:
    key_group = parser.add_mutually_exclusive_group(required=True)
    key_group.add_argument(
        "--key-location",
        metavar="FILE",
        type=str,
        help="Plaintext file in which the nova.astrometry API key is stored",
    )
    key_group.add_argument(
        "--key",
        type=str,
        help="Specify the nova.astrometry API key. NOTE: that the safer option is '--key-location'",
    )

    # Settings
    scale_group = parser.add_argument_group(
        title="Settings",
        description="Some options for basic settings / hints to give to Astrometry.net",
    )
    scale_group.add_argument(
        "--plate-scale",
        metavar="FLOAT",
        type=float,
        help="Unit: arcsec / pixel. Give an estimate for the plate scale of your instrument. Astrometry.net will look around: plate-scale within plate-scale-err of that value",
    )
    scale_group.add_argument(
        "--plate-scale-err",
        metavar="FLOAT",
        type=float,
        default=10,
        help="Specify the range (percent) around the 'plate-scale' value for Astrometry.net to search for. Default=10 percent",
    )
    scale_group.add_argument(
        "--plate-scale-range",
        metavar="FLOAT",
        nargs=2,
        type=float,
        help="Unit: arcsec / pixel. Same as 'plate-scale' but instead gives the range explicitiy (LOWER, UPPER)",
    )

    scale_group.add_argument(
        "--fov-width",
        metavar="FLOAT",
        type=float,
        help="Unit: arcmin. Give an estimate for the FoV width of the instrument. Astrometry.net will look around: fov-width within fov-width-err of that value.",
    )
    scale_group.add_argument(
        "--fov-width-err",
        metavar="FLOAT",
        type=float,
        default=10,
        help="Specify the range (percent) around the 'fov-width' value for Astrometry.net to search for. Default=10 percent",
    )
    scale_group.add_argument(
        "--fov-width-range",
        metavar="FLOAT",
        nargs=2,
        type=float,
        help="Unit: arcmin. Same as 'fov-width' but instead gives the range explicitiy (LOWER, UPPER)",
    )

    return parser.parse_args()


if __name__ == "__main__":
    main()



import os
from unittest import mock

import pytest
import requests
from constants import VALID_KEY
from utils import FunctionCalledException, function_called_raiser

from astrometry_net_client import Session
from astrometry_net_client.exceptions import APIKeyError, LoginFailedException

some_key = "somekey"


# Start of tests
def test_session_key_input_invalid():
    with pytest.raises(APIKeyError):
        Session()


def test_session_key_input_string():
    s = Session(some_key)
    assert not s.logged_in
    assert s.api_key == some_key


def test_session_key_input_file():
    s = Session(key_location="./tests/data/testkey")
    assert not s.logged_in
    assert s.api_key == some_key


@mock.patch.dict(os.environ, {"ASTROMETRY_API_KEY": some_key})
def test_session_key_input_environment():
    s = Session()
    assert not s.logged_in
    assert s.api_key == some_key


def test_valid_session_login(mock_server, monkeypatch):
    session = Session(api_key=VALID_KEY)
    session.login()  # login for the first time
    assert session.logged_in
    assert getattr(session, "key", None)  # token exists
    original_key = session.key

    # We patch the post call to send an error if it is called.
    monkeypatch.setattr(requests, "post", function_called_raiser)

    session.login()  # login should not be done now, as it is already done
    assert session.logged_in
    assert session.key == original_key

    # Here we force the login which should raise the patched exception
    with pytest.raises(FunctionCalledException):
        session.login(force=True)


def test_invalid_session_login(mock_server):
    session = Session(api_key="invalid_key")
    with pytest.raises(LoginFailedException):
        session.login()



import pytest
from constants import FILE, URL_TO_UPLOAD, VALID_KEY

from astrometry_net_client import FileUpload, Session
from astrometry_net_client.uploads import URLUpload


def test_upload(mock_server):
    session = Session(VALID_KEY)

    upl = FileUpload(FILE, session)
    assert upl.filename == FILE

    submission = upl.submit()

    assert submission.id == 2
    assert not submission.done()
    assert not submission.success()

    submission.status()

    assert not submission.done()
    assert not submission.success()

    submission.until_done()

    assert submission.done()
    job = submission.jobs[0]

    assert job.id == 2
    assert not job.done()

    job.until_done()

    assert job.done()
    assert job.success()
    assert submission.success()


def test_url_upload(mock_server):
    session = Session(VALID_KEY)

    upl = URLUpload(URL_TO_UPLOAD, session)

    assert upl.url_to_upload == URL_TO_UPLOAD

    submission = upl.submit()

    assert submission.id == 2
    assert not submission.done()
    assert not submission.success()

    submission.status()

    assert not submission.done()
    assert not submission.success()

    submission.until_done()

    assert submission.done()
    job = submission.jobs[0]

    assert job.id == 2
    assert not job.done()

    job.until_done()

    assert job.done()
    assert job.success()
    assert submission.success()


@pytest.mark.long
def test_upload_until_done(mock_server):
    session = Session(VALID_KEY)
    upl = FileUpload(FILE, session)
    submission = upl.submit()
    submission.until_done()
    assert submission.done()



import logging
from sys import argv

from astropy.io.fits import Header

from astrometry_net_client.session import Session
from astrometry_net_client.settings import Settings
from astrometry_net_client.statusables import Submission
from astrometry_net_client.uploads import FileUpload

path = "/home/sten/Documents/Projects/PracticalAstronomyCrew"
key_location = path + "/key"
filename = path + "/test-data/target.200417.00000088.3x3.FR.fits"


class DecreasedOutputFormatter(logging.Formatter):
    def format(self, record):
        result = super().format(record)
        msg = result
        if len(msg) > 300:
            result = msg[:100] + " ... " + msg[-20:]

        return result


log = logging.getLogger("Testing")


def setup_logging(full=False):
    # Setup logging
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter if full else DecreasedOutputFormatter

    # create formatter
    formatter = formatter("%(name)s - %(levelname)s - %(message)s")

    # add formatter to ch
    ch.setFormatter(formatter)

    # add ch to logger
    logging.basicConfig(handlers=[ch], level=logging.DEBUG)


def login():
    log.debug("Creating session")
    s = Session(None, key_location=key_location)

    log.debug("Loggin in")
    s.login()

    log.debug(s)
    log.debug(s.key)
    return s


def main():
    settings = Settings()
    settings.set_scale_range(10, 20)
    log.debug("Using settings: {}".format(settings))

    if "online" in argv[1:]:

        session = login()
        upl = FileUpload(filename, session=session, settings=settings)

        log.debug(upl)
        log.debug("Uploading")
        submission = upl.submit()

        log.debug("Got submission: {}".format(submission))
        log.debug("id {}".format(submission.id))
    else:
        submission = Submission(3723552)

    log.debug("Before status: {}".format(submission))
    resp = submission.until_done()
    log.debug("After status: {}".format(submission))

    log.debug("Response: {}".format(resp))
    log.debug("Start:  {}".format(submission.processing_started))
    log.debug("User Img {}".format(submission.user_images))
    log.debug("Calibrations {}".format(submission.job_calibrations))
    log.debug("Jobs {}".format(submission.jobs))

    for job in submission:
        log.debug("Before status: {}".format(job))
        resp = job.until_done()
        log.debug("After Status: {}".format(job))

        resp = job.info()
        log.debug("Info: {}".format(resp))
        # TODO: make sure to only do this when job is successfull
        # TODO:  decide how to handle... exception by wrapping around
        # TODO:  invalid request of this type will result in response 505
        # TODO: Check the response status at request level.
        wcs = job.wcs_file()
        log.debug("WCS is Header: {}".format(isinstance(wcs, Header)))


if __name__ == "__main__":
    full_logging = "full" in argv[1:]

    setup_logging(full_logging)
    main()



import pytest

from astrometry_net_client import Settings


def test_dict():
    """
    Test normal behaviour of the settings object, using multiple access
    mechanisms.
    """
    # Creating 'new' settings
    settings = Settings(parity=2)
    settings.image_width = 1000
    settings["image_height"] = 2000
    settings.update({"publicly_visible": "y", "use_sextractor": True})

    assert settings == {
        "publicly_visible": "y",
        "use_sextractor": True,
        "parity": 2,
        "image_width": 1000,
        "image_height": 2000,
    }, "Initialization incorrect"

    # Modification of the settings object
    del settings["parity"]
    settings.use_sextractor = False
    settings["publicly_visible"] = "n"
    settings.update({"image_height": 500, "image_width": 1000})

    assert settings == {
        "publicly_visible": "n",
        "use_sextractor": False,
        "image_width": 1000,
        "image_height": 500,
    }, "Modifications or deletions not applied properly"


def test_range():
    """
    Test the range convenience functions
    """
    settings = Settings()
    settings.set_scale_range(10, 20)
    assert settings == {
        "scale_lower": 10,
        "scale_upper": 20,
        "scale_units": "arcminwidth",
        "scale_type": "ul",
    }

    settings2 = Settings()
    settings2.set_scale_estimate(15, 5)
    assert settings2 == {
        "scale_est": 15,
        "scale_err": 5,
        "scale_units": "arcminwidth",
        "scale_type": "ev",
    }

    settings2 = Settings()
    settings2.set_scale_estimate(150, 5, unit="degwidth")
    assert settings2 == {
        "scale_est": 150,
        "scale_err": 5,
        "scale_units": "degwidth",
        "scale_type": "ev",
    }


def test_validation():
    """
    Tests all the validation mechanisms, where there should be an error msg
    """
    settings = Settings()

    # not a valid name for a setting
    with pytest.raises(KeyError) as e:
        settings.invalid_name = 10
    assert "invalid_name is not an allowed setting." in str(e.value)

    # valid name, but invalid type
    with pytest.raises(TypeError) as e:
        settings.scale_lower = "a string"  # expects float
    assert "Value does not have the correct type" in str(e.value)

    # valid name, but invalid type
    with pytest.raises(TypeError) as e:
        settings.image_height = 10.5  # float, expects int
    assert "Value does not have the correct type" in str(e.value)

    with pytest.raises(TypeError) as e:
        settings.allow_modifications = 0  # expects str
    assert "Value does not have the correct type" in str(e.value)

    # valid name, valid type, invalid value
    with pytest.raises(ValueError) as e:
        settings.allow_commercial_use = "invalid"
    assert "Value invalid is not allowed for the setting" in str(e.value)

    # valid name, valid type, invalid value
    with pytest.raises(ValueError) as e:
        settings.set_scale_range(0, 20, unit="invalid")
    assert "Value invalid is not allowed for the setting" in str(e.value)

    with pytest.raises(TypeError) as e:
        settings.set_scale_range(0, "invalid")
    assert "Value does not have the correct type" in str(e.value)

    with pytest.raises(TypeError) as e:
        settings.set_scale_range("invalid", 20)
    assert "Value does not have the correct type" in str(e.value)

    # make sure nothing is actually set!
    assert settings == {}, "Settings dict is not empty, but should be."


def test_data_override():
    """
    Make sure the data attribute is still accessible and does not give an error
    """
    settings = Settings()
    assert isinstance(settings.data, dict)



# Tests concerning Statusables
import pytest
from constants import FAILED_SUBMISSION_RESULT, SUCCESS_SUBMISSION_RESULT

from astrometry_net_client import Job, Submission
from astrometry_net_client.exceptions import StillProcessingException


@pytest.mark.parametrize(
    "statusable, success",
    [
        (Submission(4113880), False),
        (Submission(3781056), True),
        (Job(4450465), False),
        (Job(4446851), True),
    ],
)
@pytest.mark.online
def test_statusables(statusable, success):
    # No status queried yet
    assert not statusable.done()
    assert not statusable.success()

    resp = statusable.status()

    assert resp is not None
    assert statusable.done()
    assert statusable.success() == success


@pytest.mark.parametrize(
    "sid, result",
    [
        (4113880, FAILED_SUBMISSION_RESULT),
        (3781056, SUCCESS_SUBMISSION_RESULT),
    ],
)
@pytest.mark.online
def test_submission_response(sid, result):
    submission = Submission(sid)

    response = submission.status()
    ud_response = submission.until_done()

    assert response == result
    assert response == ud_response

    for key in result.keys():
        print(f"key = {key}")
        assert hasattr(submission, key)
        if key != "jobs":
            assert getattr(submission, key) == result[key]

    # test iteration is done properly
    for j1, j2 in zip(submission, submission.jobs):
        print(f"j1 = {j1}, j2 = {j2}")
        assert j1 == j2

    # test if job ids correspond correctly
    for job in submission:
        assert job.id in result["jobs"]


@pytest.mark.parametrize(
    "sid, success, result",
    [
        (0, False, FAILED_SUBMISSION_RESULT),
        (1, True, SUCCESS_SUBMISSION_RESULT),
    ],
)
def test_submission(sid, success, result, mock_server):
    submission = Submission(sid)

    strsubm = str(submission)
    reprsubm = repr(submission)
    assert strsubm != ""
    assert reprsubm != ""

    assert not submission.success()
    assert not submission.done()

    response = submission.status()

    assert submission.success() == success
    assert submission.done()

    assert str(submission) != strsubm
    assert repr(submission) == reprsubm

    # Make sure the until done function produces the same result
    ud_response = submission.until_done()

    assert response == result
    assert response == ud_response

    for key in result.keys():
        print(f"key = {key}")
        assert hasattr(submission, key)
        if key != "jobs":
            assert getattr(submission, key) == result[key]

    # test iteration is done properly
    for j1, j2 in zip(submission, submission.jobs):
        print(f"j1= {j1}, j2 = {j2}")
        assert j1 == j2

    # test if job ids correspond correctly
    for job in submission:
        assert job.id in result["jobs"]


def test_submission_still_solving(mock_server):
    submission = Submission(2)
    with pytest.raises(StillProcessingException):
        iter(submission)



import os
from unittest import mock

import pytest
from constants import FILE, VALID_KEY

from astrometry_net_client import Client, Session, Settings
from astrometry_net_client.exceptions import LoginFailedException

some_key = "somekey"


@mock.patch.dict(os.environ, {"ASTROMETRY_API_KEY": VALID_KEY})
def test_client_login_env(mock_server):
    c = Client()
    assert c.session.logged_in


def test_client_login(mock_server):
    c = Client(api_key=VALID_KEY)
    assert c.session.logged_in

    with pytest.raises(LoginFailedException):
        c = Client(key_location="./tests/data/testkey")

    c = Client(session=Session(VALID_KEY))
    assert c.session.logged_in


def test_client_settings(mock_server):
    settings = Settings(use_sextractor=True, image_height=1000, image_width=2000)
    client = Client(settings=settings, api_key=VALID_KEY)
    client2 = Client(
        use_sextractor=True,
        image_height=1000,
        image_width=2000,
        api_key=VALID_KEY,
    )
    assert client.settings == {
        "use_sextractor": True,
        "image_height": 1000,
        "image_width": 2000,
    }
    assert client.settings == client2.settings


@pytest.mark.long
def test_client_upload(mock_server):
    client = Client(api_key=VALID_KEY)

    job = client.upload_file(FILE)
    assert job.success()
    assert job.done()


@pytest.mark.long
def test_client_upload_multiple(mock_server):
    client = Client(api_key=VALID_KEY)
    jobs = client.upload_files_gen([FILE] * 5, queue_size=3)

    assert jobs is not None

    for job, filename in jobs:
        assert filename == FILE
        assert job.done()
        assert job.success()



import pytest
import requests
from utils import function_called_raiser

from astrometry_net_client import Job
from astrometry_net_client.exceptions import (
    StatusFailedException,
    StillProcessingException,
)


@pytest.mark.mocked
def test_mocked_job_status_success(mock_server, monkeypatch):
    job = Job(1)

    strjob = str(job)
    reprjob = repr(job)
    assert strjob != ""
    assert reprjob != ""

    assert hash(job) == hash(1)
    # No status queried yet
    assert not job.done()
    assert not job.success()

    response = job.status()

    assert job.done()
    assert job.success()
    assert response == {"status": "success"}
    assert job.resp_status == "success"

    assert str(job) != strjob
    assert repr(job) == reprjob

    info = job.info()
    assert info is not None
    assert isinstance(info, dict)
    assert info["status"] == "success"

    for key in info.keys():
        print(f"key = {key}")
        assert hasattr(job, key)
        if key != "status":
            assert getattr(job, key) == info[key]

    # Test caching of info(). If get is used, throw exception.
    monkeypatch.setattr(requests, "get", function_called_raiser)

    # No new request should be made!
    cached_info = job.info()
    assert cached_info is info


@pytest.mark.mocked
def test_mocked_job_status_failure(mock_server):
    job = Job(0)

    # No status queried yet
    assert not job.done()
    assert not job.success()

    response = job.status()

    assert response == {"status": "failure"}
    assert job.resp_status == "failure"
    assert job.done()
    assert not job.success()

    info = job.info()
    assert info is not None
    assert info.get("calibration") is None
    assert isinstance(info, dict)
    assert info["status"] == "failure"

    for key in info.keys():
        print(f"{key}")
        assert hasattr(job, key)
        if key != "status":
            assert getattr(job, key) == info[key]

    # Should have this exception
    with pytest.raises(StatusFailedException):
        job.wcs_file()


def test_job_still_solving(mock_server):
    job = Job(2)

    with pytest.raises(StillProcessingException):
        job.info()


@pytest.mark.long
def test_job_timeout(mock_server):
    job = Job(2)
    with pytest.raises(TimeoutError):
        job.until_done(timeout=0)  # immediate timeout

    job_2 = Job(3)
    with pytest.raises(TimeoutError):
        # timeout after two queries
        job_2.until_done(start=1, end=1, timeout=2)



import pytest
from astropy.io.fits import Header

from astrometry_net_client import FileUpload, Job, Session, Submission
from astrometry_net_client.exceptions import StatusFailedException


@pytest.mark.online
def test_job_status_success(success_job: Job):
    response = success_job.status()
    assert success_job.done()
    assert success_job.success()
    assert response == {"status": "success"}
    assert success_job.resp_status == "success"

    wcs = success_job.wcs_file()
    assert isinstance(wcs, Header)


@pytest.mark.online
def test_job_status_failed(failed_job: Job):
    response = failed_job.status()

    assert failed_job.done()
    assert not failed_job.success()
    assert response == {"status": "failure"}
    assert failed_job.resp_status == "failure"

    with pytest.raises(StatusFailedException):
        failed_job.wcs_file()


@pytest.mark.online
@pytest.mark.long
def test_old():
    # TODO: verify if this function is still correct & all tests pass
    # TODO: does pass, but no checks done for consistency

    # WARNING: makes an upload and then waits for the upload to be done.
    #          do not spam this test!

    # key in project root
    session = Session(None, key_location="./key")
    assert not session.logged_in

    session.login()
    assert session.logged_in

    # TODO include some test data in the repository
    upl = FileUpload("../test/data/target.200417.00000088.3x3.FR.fits", session=session)

    submission = upl.submit()
    assert isinstance(submission, Submission)

    assert not submission.done()
    resp = submission.status()
    assert not submission.done()

    resp = submission.until_done()
    assert submission.done()

    status_attrs = [
        "processing_started",
        "processing_finished",
        "user_images",
        "user",
        "images",
        "job_calibrations",
    ]
    for attr in status_attrs:
        assert hasattr(submission, attr), f"Submission does not have {attr}"

    assert len(submission.jobs) > 0
    for job in submission:
        resp = job.until_done()
        assert job.done()
        assert job.success()

        resp = job.info()
        assert resp["status"] == "success"
        assert job.resp_status == "success"
        assert job.calibration is not None

        wcs = job.wcs_file()
        assert isinstance(wcs, Header)
